import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision
import torchvision.transforms as transforms
import torch.optim as optim
from PIL import Image
from torchvision.utils import make_grid

import matplotlib.pyplot as plt
import numpy as np

# https://papers.nips.cc/paper_files/paper/2012/hash/c399862d3b9d6b76c8436e924a68c45b-Abstract.html

class Net(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(in_channels=1, out_channels=16, kernel_size=5, stride=1, padding=2)
        self.conv2 = nn.Conv2d(16, 32, 5, 1, 2)
        self.conv3 = nn.Conv2d(32, 48, 3, 1, 1)
        #self.conv4 = nn.Conv2d(64, 128, 3, 1, 1)
        self.pool = nn.MaxPool2d(2, 2)
        self.fc1 = nn.LazyLinear(120)
        self.fc2 = nn.LazyLinear(80)
        self.fc3 = nn.LazyLinear(43)

    def forward(self, x):
        self.feature_map1 = self.pool(F.relu(self.conv1(x)))
        self.feature_map2 = self.pool(F.relu(self.conv2(self.feature_map1)))
        self.feature_map3 = self.pool(F.relu(self.conv3(self.feature_map2)))
        x = torch.flatten(self.feature_map3, 1) # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x

class MyAlexNet(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(in_channels=3, out_channels=96, kernel_size=11, stride=4)
        self.conv2 = nn.Conv2d(in_channels=96, out_channels=256, kernel_size=5, stride=1, padding=2)
        self.conv3 = nn.Conv2d(in_channels=256, out_channels=384, kernel_size=3, stride=1, padding=1)       
        self.conv4 = nn.Conv2d(in_channels=384, out_channels=384, kernel_size=3, stride=1, padding=1)        
        self.conv5 = nn.Conv2d(in_channels=384, out_channels=256, kernel_size=3, stride=1, padding=1)        
        
        #self.pool = nn.MaxPool2d(2, 2)
        self.pool = nn.MaxPool2d(kernel_size = 3, stride = 2)
        self.fc1 = nn.LazyLinear(4096)#nn.Linear(9216,4096)
        self.fc2 = nn.LazyLinear(4096)
        self.fc3 = nn.LazyLinear(43)

    def forward(self, x):
        self.feature_map1 = self.pool(F.relu(self.conv1(x)))
        self.feature_map2 = self.pool(F.relu(self.conv2(self.feature_map1)))
        self.feature_map3 = F.relu(self.conv3(self.feature_map2))
        self.feature_map4 = F.relu(self.conv4(self.feature_map3))
        self.feature_map5 = self.pool(F.relu(self.conv5(self.feature_map4)))
        x = torch.flatten(self.feature_map5, 1) # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


net = MyAlexNet() #torchvision.models.alexnet(pretrained=True)     

from torchsummary import summary
summary(net, (3, 227, 227), device="cpu")

torch_alex = torchvision.models.alexnet(weights=None)
print(torch_alex)
torch_alex.classifier[6] = nn.Linear(4096, 43)
print(torch_alex)
summary(torch_alex, (3, 227, 227), device="cpu")
